# -*- coding: utf-8 -*-

"""
This middleware is copy/pasted from the following post:
    https://stackoverflow.com/questions/23398393/
    (note as of 8.03.2023, this post has bugs in it which I fixed below)

We may switch to an alternative fix in the future:
    https://stackoverflow.com/questions/48407790/
"""

import re

from django.conf import settings
from django.contrib.auth.decorators import login_required
from rest_framework.authentication import TokenAuthentication


class RequireLoginMiddleware:
    """
    Middleware component that wraps the login_required decorator around
    matching URL patterns. To use, add the class to MIDDLEWARE_CLASSES and
    define LOGIN_REQUIRED_URLS and LOGIN_REQUIRED_URLS_EXCEPTIONS in your
    settings.py. For example:
    ------
    LOGIN_REQUIRED_URLS = (
        r'/topsecret/(.*)$',
    )
    LOGIN_REQUIRED_URLS_EXCEPTIONS = (
        r'/topsecret/login(.*)$',
        r'/topsecret/logout(.*)$',
    )
    ------
    LOGIN_REQUIRED_URLS is where you define URL patterns; each pattern must
    be a valid regex.

    LOGIN_REQUIRED_URLS_EXCEPTIONS is, conversely, where you explicitly
    define any exceptions (like login and logout URLs).
    """

    def __init__(self, get_response):
        self.get_response = get_response
        self.required = tuple(re.compile(url) for url in settings.LOGIN_REQUIRED_URLS)
        self.exceptions = tuple(
            re.compile(url) for url in settings.LOGIN_REQUIRED_URLS_EXCEPTIONS
        )

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):

        # No need to process URLs if user already logged in
        if request.user.is_authenticated:
            return

        # Users can also authenticate via Tokens from Django REST Framework
        # for programmatic access
        token_auth = TokenAuthentication()
        try:
            # Returns None if not signed-in OR if no token was given
            # Returns (user, token) if successful
            # Raises an error if the token is invalid or improperly formatted
            # TODO: redirect to helpful pages for failed token auth
            if token_auth.authenticate(request):
                return

        except:  # any error corresponds to not authenticated
            pass

        # An exception match should immediately return None
        for url in self.exceptions:
            if url.match(request.path):
                return

        # Requests matching a restricted URL pattern are returned
        # wrapped with the login_required decorator
        for url in self.required:
            if url.match(request.path):
                return login_required(view_func)(request, *view_args, **view_kwargs)

        # Explicitly return None for all non-matching requests
        return
